* Adapted from code written by Cetin Basoz, http://www.tek-tips.com/viewthread.cfm?qid=1605291

* Test code for the ExcelExportAutomation class

open database (_samples + 'data\testdata')
use Employee
loExport = createobject('ExcelExportAutomation')
loExport.lShowExcel = .T.
if not loExport.Export('EmployeeOLE.xlsx')
	messagebox(loExport.cErrorMessage)
endif not loExport.Export('EmployeeOLE.xlsx')

* This class exports to Excel using OLE Automation.

define class ExcelExportAutomation as Custom
	lShowExcel       = .F.
		&& .T. to leave the document open and Excel visible
	lIncludeHeadings = .T.
		&& .T. to put field names in the first row
	cErrorMessage    = ''
		&& the text of any error that occurs

* Exports the table in the current workarea to the specified Excel file.

	function Export(tcFileName)
		local loExcel, ;
			loException, ;
			lcDatabase, ;
			lcDataSource, ;
			lcTable, ;
			loSheet, ;
			llReturn
		try
			loExcel = createobject('Excel.Application')
		catch to loException
			This.cErrorMessage = 'Could not instantiate Excel. ' + ;
				'The error message is: ' + loException.Message
		endtry
		if vartype(loExcel) = 'O'
			with loExcel

* Determine the datasource (the DBC if there is one or the folder for the
* current table if not) and SQL statement (just USE TABLE in this case) to use.

				lcDatabase = cursorgetprop('Database')
				if empty(lcDatabase)
					lcDataSource = '"' + justpath(dbf()) + '"'
				else
					lcDataSource = '"' + lcDatabase + '"'
				endif empty(lcDatabase)
				lcTable = 'use "' + dbf() + '"'

* Create a workbook and output the data to it, then save the file.

				.WorkBooks.Add()
				loSheet  = .ActiveWorkBook.ActiveSheet
				llReturn = This.VFP2Excel(lcDataSource, lcTable, ;
					loSheet.Range('A1'))
				if llReturn
					loSheet.UsedRange.Columns.Autofit()
					llReturn = This.SaveDocument(loExcel, .ActiveWorkBook, ;
						tcFileName)
				endif llReturn

* Either display Excel or close it.

				if This.lShowExcel and llReturn
					.Visible = .T.
				else
					.Quit()
				endif This.lShowExcel ...
			endwith
		endif vartype(loExcel) = 'O'
		return llReturn
	endfunc

* Fills the specified range with the data from the specified SQL statement or
* table.

	protected function VFP2Excel(tcDataSource, tcSQL, toRange)
		local loConn, ;
			loRS, ;
			lnI, ;
			lnRow, ;
			llReturn
		try
			loConn = createobject('ADODB.Connection')
			loConn.ConnectionString = 'Provider=VFPOLEDB;Data Source=' + ;
				tcDataSource
			loConn.Open()
			loRS = loConn.Execute(tcSQL)
			if This.lIncludeHeadings
				for lnI = 0 to loRS.Fields.Count - 1
					toRange.Offset(0, lnI).Value = proper(loRS.Fields(lnI).Name)
					toRange.Offset(0, lnI).Font.Bold = .T.
				next lnI
				lnRow = 1
			else
				lnRow = 0
			endif This.lIncludeHeadings
			toRange.Offset(lnRow, 0).CopyFromRecordSet(loRS)
			loRS.Close()
			loConn.Close()
			llReturn = .T.
		catch to loException
			This.cErrorMessage = 'Could not output data to workbook. The ' + ;
				'error message is: '  + alltrim(substr(loException.Message, ;
				at(':', loException.Message) + 1))
		endtry
		return llReturn
	endfunc

* Saves the workbook to the specified file.

	protected function SaveDocument(toExcel, toWorkbook, tcFileName)
		local llAlerts, ;
			lcFileName, ;
			llReturn

* Set DisplayAlerts to .F. to avoid any dialogs.

		try
			llAlerts = toExcel.DisplayAlerts
			toExcel.DisplayAlerts = .F.
		catch
		endtry

* Save the workbook.

		lcFileName = GetFullPath(tcFileName)
		try
			toWorkbook.SaveAs(lcFileName)
			llReturn = file(lcFileName)
		catch to loException
		endtry
		if not llReturn
			This.cErrorMessage = 'Could not save ' + lcFileName + ;
				iif(vartype(loException) = 'O', '. The error message is: ' + ;
				alltrim(substr(loException.Message, ;
				at(':', loException.Message) + 1)), '')
		endif not llReturn

* Restore DisplayAlerts.

		try
			toExcel.DisplayAlerts = llAlerts
		catch
		endtry
		return llReturn
	endfunc
enddefine
 